from pyxllib.xl import distribute_package

distribute_package('D:/slns/pyxllib', '0.1.21')

# 注意这个版本机制不太一样，ui界面标题的版本是另设的
# distribute_package('D:/slns/xllabelme', '1.8')

# distribute_package('D:/slns/xlcocotools', '0.0.3')  # 还未上传，待上传

# distribute_package('D:/slns/pyxlpr', '0.0.45', 'xlpr')
